c****************subroutine QQAAWJ *************************************
c
c	giuseppe bozzi, <giuseppe@particle.uni-karlsruhe.de>
c	Last modified: December 2009
C
C  QQVAAj calculates the matrix elements**2 for 
C     q1 q2    ---->   W+- A A g,   
C     W+- ---> f1-bar f2
C
C  and crossing related processes. Pauli interference terms for
c  identical fermions are neglected.
c
c  Gluon polarization vectors are given in the kartesian basis
c  (comparison with madgraph amplitudes tampr requires to express these in
c   kart. basis also, even though they are given in the helicity basis
c   a priori).
c
c  This code includes only real emission contributions, i.e.
c
c      return udwwz = |M_real|^2   etc.
c
c       fpials is attached only in the end of the code
c
c       l is the gluon polarization in the kartesian basis (l=1,2)
c       k is the process ID (1:uuwwaj,2:ddwwaj)
c       isig is the helicity of partons 1 and 2
c
c*********************************************************************

      subroutine qqAAWj(pbar,sign, qbar, gsign, final, bos)

      implicit none

#include "VBFNLO/utilities/global.inc"
#include "tensor.inc"
#include "VBFNLO/utilities/scales.inc"
#include "VBFNLO/utilities/coupl.inc"

c     electroweak couplings are taken from KOPPLN

      double precision  clr, xm2, xmg, b, v, a
      COMMON /BKOPOU/   CLR(4,5,-1:1),XM2(6),XMG(6),B(6,6,6),
     1                  V(4,5),A(4,5)

c variables for the main part of the program      

      DOUBLE PRECISION fpials(2:3), fpi
      PARAMETER (fpi=4d0*pi)

      double precision  pbar(0:3,2+n_v), qbar(0:4), q(0:3)
      double precision  uuwwaj, ddwwaj, res(2), final
      double precision  p(0:3,2+n_v),fq0(0:4,2),fqa1(0:4,2),fqa2(0:4,2)

      integer  sign(2+n_v), gsign, mu, i, j, k, isig
      integer  ifl(2), lh, l, bos

      double precision fqm(0:4,2), fqp(0:4,2), fqz(0:4,2)
      double precision pq(0:4,2), pgZ(0:4,2), pZg(0:4,2)
      double precision pgA1(0:4,2), pA1g(0:4,2)
      double precision pgA2(0:4,2), pA2g(0:4,2)
      double precision pgw0(0:4,2), pw0g(0:4,2)
      double precision pgWP(0:4,2), pgWM(0:4,2), pWPg(0:4,2)
      double precision pWMg(0:4,2), pwpz(0:3), pwmz(0:3) 
      double precision pwa1(0:4), pwa2(0:4) 
      double precision eps(0:3,2) ! g in kartesian basis
      double complex psi(2,-1:1,2), psiWM(2,-1:1,2), psiWP(2,-1:1,2)
      double complex psia(2,-1:1,2), psiz(2,-1:1,2)
      double complex psia1g(2,-1:1,2,2), psia2g(2,-1:1,2,2)
      double complex psiw0g(2,-1:1,2,2)
      double complex braketg(2,-1:1,2,2)
      double complex braketgw0(2,-1:1,2,2)
      double complex braketgWP(2,-1:1,2,2), braketgWM(2,-1:1,2,2)
      double complex braketgA(2,-1:1,2,2), braketgZ(2,-1:1,2,2)
      double complex braketgA1(2,-1:1,2,2), braketgA2(2,-1:1,2,2)
      double complex psiWPg(2,-1:1, 2,2), psiWMg(2,-1:1, 2,2)
      double complex psiw0(2,-1:1, 2),psia1(2,-1:1, 2),
     1               psia2(2,-1:1, 2)
      double complex psiAg(2,-1:1,2,2), psiZg(2,-1:1,2,2)
      double complex mat(2,-1:1,2,3), madgr(3,-1:1,-1:1,2)
      double complex matt(2,-1:1,2,3)
      double complex mm(2, -1:1,2)
      double complex m1kb(2,-1:1,3), m2kb(2,-1:1,3), temp, temp1
      double complex m1(3,2), m2(3,2), m3(3,2), m4(3,2)
      double complex ma, mz
      double complex janew(0:5,-1:1,14,2)
      double complex ja(0:5,-1:1,4,2), jz(0:5,-1:1,4,2)
      double complex jwm(0:5,-1:1,4,2), jwp(0:5,-1:1,4,2)
      double complex jqgaq(0:5,-1:1,2), jqagq(0:5,-1:1,2)
      double complex jqqa(0:5,-1:1,2)
   

      double complex im
      double complex dotcc, s1c, dotrc
      double complex zero 
      external dotcc, s1c, dotrc
      integer u(2),d(2)
      save ifl,u,k,d
      parameter (im = (0d0,1d0))
      parameter (zero = (0d0,0d0) )

c variable for q_mu q_nu terms in the propagators of the massive vector bosons

      double complex zm2i(2:3)
      save zm2i

c  helicity selection

      INTEGER h
      COMMON /hcount / h

      logical ldebug, ldebugm, linit
      data linit /.true./
      save linit
      parameter(ldebugm = .false., ldebug=.false.)

c -------------- BEGIN CODE --------------------------------

c initialize & precompute stuff needed below:

c fix strong coupling gs**2 for the two quarks:

      fpials(2) = fpi*als(1,1)
      fpials(3) = fpi*als(2,1)
c      fpials(2) = fpi*alfas

c define flavors of external quarks for the 2 NC subprocesses

      if (linit) then
         linit = .false.
         if (bos.eq.311) then
            k = 1               ! udwaaj
            u(1) = 3
            d(1) = 4 
         elseif (bos.eq.411) then
            k = 2               ! duwwaj
            u(2) = 4
            d(2) = 3            
         endif
         zm2i(3) = 1d0/dcmplx(xm2(3),-xmg(3))
      endif

      do isig = -1,1,2
         do l = 1,2
            do i = 1,3
               mat(k,isig,l,i) = (0.0d0, 0.0d0)
            enddo
         enddo
      enddo

c define the internal momenta

      do mu = 0,3
         do i = 1,2+n_v
            p(mu,i) = pbar(mu,i)*sign(i)
         enddo
	 q(mu) = qbar(mu)*gsign
         pwa1(mu) = qw0(mu,1) + qa1(mu,1)
         pwa2(mu) = qw0(mu,1) + qa2(mu,1)
      enddo
      pwa1(4) = pwa1(0)**2-pwa1(1)**2-pwa1(2)**2-pwa1(3)**2
      pwa2(4) = pwa2(0)**2-pwa2(1)**2-pwa2(2)**2-pwa2(3)**2

c get the external quark spinors (including factor sqrt(2E) )

      call psi0m(2,pbar(0,1),sign(1),psi)

c get other bra and ket vectors

      call bra2c(psi(1,-1,2),.true.,p(0,2),-1,qw0(0,1),w0(1,1),
     &              psiw0(1,-1,2),fq0(0,2))
      call ket2c(psi(1,-1,1),.true.,p(0,1),-1,qw0(0,1),w0(1,1),
     &              psiw0(1,-1,1),fq0(0,1))
      call bra2c(psi(1,-1,2),.true.,p(0,2),-1,qa1(0,1),
     &     atau1(1,1),psia1(1,-1,2),fqa1(0,2))
      call ket2c(psi(1,-1,1),.true.,p(0,1),-1,qa1(0,1),
     &     atau1(1,1),psia1(1,-1,1),fqa1(0,1))
      call bra2c(psi(1,-1,2),.true.,p(0,2),-1,qa2(0,1),
     &     atau2(1,1),psia2(1,-1,2),fqa2(0,2))
      call ket2c(psi(1,-1,1),.true.,p(0,1),-1,qa2(0,1),
     &     atau2(1,1),psia2(1,-1,1),fqa2(0,1))
      
c  Get the gluon polarization vector and the gluon emission spinors

      do l = 1,2                ! 2 gluon polarizations

         call polvec(qbar,l,eps(0,l)) ! get gluon pol.vectors

c---- QCD gaugecheck --- eps_mu -> k_mu --- k_mu M^mu = 0 ------------

c         do mu = 0,3                   !QCD gaugecheck      
c            eps(mu,l) = qbar(mu)
c         enddo

c--- End of QCD gaugecheck -------------------------------------------    

c     NOTES for bras and kets: .true. if psi is a 2-spinor of the chi
c     form as output by psi0m, .false. otherwise.  the last entry is
c     the sum of the two momenta (p plus q) and effectively the
c     momentum of the new spinor.
 
         call ket2r(psi(1,-1,1),.true.,p(0,1),-1,q,eps(0,l),
     &        braketg(1,-1,1,l),pq(0,1)) ! |q,1>_l,isig
         
         call bra2r(psi(1,-1,2),.true.,p(0,2),-1,q,eps(0,l),
     &        braketg(1,-1,2,l),pq(0,2)) ! <2,q|_l,isig2
         
c     braketg contains the free quark spinors multiplied by a fermion
c     propagator and a gluon eps_slash. 
c     NOTATION: braketg(2 component spinor, isig =-1 or 1 (fermion hel.),
c     fermion ID = 1:4, gluon polarization l=1:2)

         call ket2c(braketg(1,-1,1,l),.false.,pq(0,1),-1,
     &        qa1(0,1),atau1(1,1),braketgA1(1,-1,1,l),pgA1(0,1))
         
         call bra2c(braketg(1,-1,2,l),.false.,pq(0,2),-1,
     &        qa1(0,1),atau1(1,1),braketgA1(1,-1,2,l),pgA1(0,2))
         
         call ket2r(psia1(1,-1,1),.false.,fqa1(0,1),-1,q,eps(0,l),
     &        psiA1g(1,-1,1,l),pA1g(0,1))
         
         call bra2r(psia1(1,-1,2),.false.,fqa1(0,2),-1,q,eps(0,l),
     &        psiA1g(1,-1,2,l),pA1g(0,2))

         call ket2c(braketg(1,-1,1,l),.false.,pq(0,1),-1,
     &        qa2(0,1),atau2(1,1),braketgA2(1,-1,1,l),pgA2(0,1))
         
         call bra2c(braketg(1,-1,2,l),.false.,pq(0,2),-1,
     &        qa2(0,1),atau2(1,1),braketgA2(1,-1,2,l),pgA2(0,2))
         
         call ket2r(psia2(1,-1,1),.false.,fqa2(0,1),-1,q,eps(0,l),
     &        psiA2g(1,-1,1,l),pA2g(0,1))
         
         call bra2r(psia2(1,-1,2),.false.,fqa2(0,2),-1,q,eps(0,l),
     &        psiA2g(1,-1,2,l),pA2g(0,2))
         
         call ket2c(braketg(1,-1,1,l),.false.,pq(0,1),-1,
     &        qw0(0,1),w0(1,1),braketgw0(1,-1,1,l),pgw0(0,1)) 

         call bra2c(braketg(1,-1,2,l),.false.,pq(0,2),-1,
     &        qw0(0,1),w0(1,1),braketgw0(1,-1,2,l),pgw0(0,2))

         call ket2r(psiw0(1,-1,1),.false.,fq0(0,1),-1,
     &        q,eps(0,l),psiw0g(1,-1,1,l),pw0g(0,1))

         call bra2r(psiw0(1,-1,2),.false.,fq0(0,2),-1,
     &        q,eps(0,l),psiw0g(1,-1,2,l),pw0g(0,2))
         
      enddo

c Calculation of processes qq -> W A A g ---------------------

c   u d -> W A A g

      do l=1,2

         m1(1,l) =
     &      -s1c(psiA1(1,-1,2),atau2(1,1),.true.,-1,braketgw0(1,-1,1,l))
     &      -s1c(psiA2(1,-1,2),atau1(1,1),.true.,-1,braketgw0(1,-1,1,l))
     &      -s1c(psiA1(1,-1,2),atau2(1,1),.true.,-1,psiw0g(1,-1,1,l))
     &      -s1c(psiA2(1,-1,2),atau1(1,1),.true.,-1,psiw0g(1,-1,1,l))
     &      -s1c(braketgA1(1,-1,2,l),atau2(1,1),.true.,-1,psiw0(1,-1,1))
     &      -s1c(braketgA2(1,-1,2,l),atau1(1,1),.true.,-1,psiw0(1,-1,1))
     &      -s1c(psiA1g(1,-1,2,l),atau2(1,1),.true.,-1,psiw0(1,-1,1))
     &      -s1c(psiA2g(1,-1,2,l),atau1(1,1),.true.,-1,psiw0(1,-1,1))

         m1(2,l) =
     &      -s1c(psiw0(1,-1,2),atau1(1,1),.true.,-1,braketgA2(1,-1,1,l))
     &      -s1c(psiw0(1,-1,2),atau2(1,1),.true.,-1,braketgA1(1,-1,1,l))
     &      -s1c(psiw0(1,-1,2),atau1(1,1),.true.,-1,psiA2g(1,-1,1,l))
     &      -s1c(psiw0(1,-1,2),atau2(1,1),.true.,-1,psiA1g(1,-1,1,l))
     &      -s1c(braketgw0(1,-1,2,l),atau1(1,1),.true.,-1,psiA2(1,-1,1))
     &      -s1c(braketgw0(1,-1,2,l),atau2(1,1),.true.,-1,psiA1(1,-1,1))
     &      -s1c(psiw0g(1,-1,2,l),atau1(1,1),.true.,-1,psiA2(1,-1,1))
     &      -s1c(psiw0g(1,-1,2,l),atau2(1,1),.true.,-1,psiA1(1,-1,1))

         m1(3,l) =
     &      -s1c(psiA1(1,-1,2),w0(1,1),.true.,-1,braketgA2(1,-1,1,l))
     &      -s1c(psiA2(1,-1,2),w0(1,1),.true.,-1,braketgA1(1,-1,1,l))
     &      -s1c(psiA1(1,-1,2),w0(1,1),.true.,-1,psiA2g(1,-1,1,l))
     &      -s1c(psiA2(1,-1,2),w0(1,1),.true.,-1,psiA1g(1,-1,1,l))
     &      -s1c(braketgA1(1,-1,2,l),w0(1,1),.true.,-1,psiA2(1,-1,1))
     &      -s1c(braketgA2(1,-1,2,l),w0(1,1),.true.,-1,psiA1(1,-1,1))
     &      -s1c(psiA1g(1,-1,2,l),w0(1,1),.true.,-1,psiA2(1,-1,1))
     &      -s1c(psiA2g(1,-1,2,l),w0(1,1),.true.,-1,psiA1(1,-1,1))

         mat(k,-1,l,1) = (m1(1,l) * clr(d(k),1,-1)**2 
     &                  + m1(2,l) * clr(u(k),1,-1)**2
     &                  + m1(3,l) * clr(u(k),1,-1)*clr(d(k),1,-1)) 
     &                  * clr(3,3,-1) 

      enddo


c Calculation of processes qq -> W A g -> W A A g -------------

c qq -> A W(-> W A) g

      do l=1,2

         call curr6(-1,psi(1,-1,2),p(0,2),
     &              braketgA1(1,-1,1,l),pgA1(0,1),janew(0,-1,1,l) )

         call curr6(-1,psi(1,-1,2),p(0,2),
     &              braketgA2(1,-1,1,l),pgA2(0,1),janew(0,-1,2,l) )

         call curr6(-1,psi(1,-1,2),p(0,2),
     &              psiA1g(1,-1,1,l),pA1g(0,1),janew(0,-1,3,l) )

         call curr6(-1,psi(1,-1,2),p(0,2),
     &              psiA2g(1,-1,1,l),pA2g(0,1),janew(0,-1,4,l) )

         call curr6(-1,braketg(1,-1,2,l),pq(0,2),
     &              psiA1(1,-1,1),fqA1(0,1),janew(0,-1,5,l) )

         call curr6(-1,braketg(1,-1,2,l),pq(0,2),
     &              psiA2(1,-1,1),fqA2(0,1),janew(0,-1,6,l) )



         call curr6(-1,psiA1(1,-1,2),fqA1(0,2),
     &              braketg(1,-1,1,l),pq(0,1),janew(0,-1,7,l) )

         call curr6(-1,psiA2(1,-1,2),fqA2(0,2),
     &              braketg(1,-1,1,l),pq(0,1),janew(0,-1,8,l) )

         call curr6(-1,psiA1g(1,-1,2,l),pA1g(0,2),
     &              psi(1,-1,1),p(0,1),janew(0,-1,9,l) )

         call curr6(-1,psiA2g(1,-1,2,l),pA2g(0,2),
     &              psi(1,-1,1),p(0,1),janew(0,-1,10,l) )

         call curr6(-1,braketgA1(1,-1,2,l),pgA1(0,2),
     &              psi(1,-1,1),p(0,1),janew(0,-1,11,l) )

         call curr6(-1,braketgA2(1,-1,2,l),pgA2(0,2),
     &              psi(1,-1,1),p(0,1),janew(0,-1,12,l) )

         do mu = 0,5
            janew(mu,-1,13,l) = janew(mu,-1,1,l) *clr(u(k),1,-1)
     &                        + janew(mu,-1,3,l) *clr(u(k),1,-1)
     &                        + janew(mu,-1,5,l) *clr(u(k),1,-1)
     &                        + janew(mu,-1,7,l) *clr(d(k),1,-1)
     &                        + janew(mu,-1,9,l) *clr(d(k),1,-1)
     &                        + janew(mu,-1,11,l)*clr(d(k),1,-1)
            janew(mu,-1,14,l) = janew(mu,-1,2,l) *clr(u(k),1,-1) 
     &                        + janew(mu,-1,4,l) *clr(u(k),1,-1)
     &                        + janew(mu,-1,6,l) *clr(u(k),1,-1)
     &                        + janew(mu,-1,8,l) *clr(d(k),1,-1)
     &                        + janew(mu,-1,10,l)*clr(d(k),1,-1)
     &                        + janew(mu,-1,12,l)*clr(d(k),1,-1)
         enddo

c contract with atoww/ztoww tensor and
c additional contributions due to q_mu q_nu/ m^2 terms in the propagator
         if (bos.eq.311) then
            m1(1,l) = -dotcc(wpwa2(0,1),janew(0,-1,13,l))
            m2(1,l) = -dotcc(wpwa1(0,1),janew(0,-1,14,l))
            m1(1,l) = m1(1,l)+zm2i(3)*dotrc(pwa2(0),wpwa2(0,1))
     1           *dotrc(pwa2(0),janew(0,-1,13,l))
            m2(1,l) = m2(1,l)+zm2i(3)*dotrc(pwa1(0),wpwa1(0,1))
     1           *dotrc(pwa1(0),janew(0,-1,14,l))
         elseif (bos.eq.411) then
            m1(1,l) = -dotcc(wmwa2(0,1),janew(0,-1,13,l))
            m2(1,l) = -dotcc(wmwa1(0,1),janew(0,-1,14,l))
            m1(1,l) = m1(1,l)+zm2i(3)*dotrc(pwa2(0),wmwa2(0,1))
     1           *dotrc(pwa2(0),janew(0,-1,13,l))
            m2(1,l) = m2(1,l)+zm2i(3)*dotrc(pwa1(0),wmwa1(0,1))
     1           *dotrc(pwa1(0),janew(0,-1,14,l))
         endif
         
         mat(k,-1,l,2) = (m1(1,l)+m2(1,l))*clr(3,3,-1)
         
      enddo                     ! gluon polarization
      

c Calculation of processes qq -> A/Z g -> W W A g ---------------------

      do l =1,2
         
         call curr6(-1,psi(1,-1,2),p(0,2),
     &        braketg(1,-1,1,l),pq(0,1),jqgaq(0,-1,l))
         
         call curr6(-1,braketg(1,-1,2,l),pq(0,2),
     &        psi(1,-1,1),p(0,1),jqagq(0,-1,l))
         
         do mu = 0,5
            jqqa(mu,-1,l) = jqgaq(mu,-1,l) + jqagq(mu,-1,l)
         enddo
         
c     contract with atowwa/ztowwa tensor
         if (bos.eq.311) then
            ma = -dotcc(wpwaa(0,1),jqqa(0,-1,l))
         elseif (bos.eq.411) then
            ma = -dotcc(wmwaa(0,1),jqqa(0,-1,l))
         endif

         mat(k,-1,l,3) = ma*clr(3,3,-1)
         
      enddo                     ! gluon polarization
      
#ifdef WITH_MADGRAPH
c------------------------------------------------------
c comparison with madgraph

      if (ldebugm) then
         
         do isig = -1,1,2
            do i = 1,3
                  
                  m1kb(1,isig,i) = 0d0 ! kartesian basis l = 1
                  m2kb(1,isig,i) = 0d0 ! kartesian basis l = 2
                  
                  do lh = -1,1,2
                     madgr(i,isig,lh,1) = 0d0
                  enddo
                  
            enddo
         enddo
         
         do isig = -1,1,2
            do lh = -1,1,2
                  do i =1,3
                     
                     madgr(i,isig,lh,1) = tampr(i,isig,lh,1)
                     
                  enddo
            enddo
         enddo
         
         do isig = -1,1,2
            do i=1,3
                  
                  m1kb(1,isig,i) = (madgr(i,isig,-1,1) - 
     &                 madgr(i,isig,1,1))/sqrt(2d0)       
                  m2kb(1,isig,i) = -im*(madgr(i,isig,-1,1) + 
     &                 madgr(i,isig,1,1))/sqrt(2d0)
                  
            enddo
         enddo
         
c print results from checks:

         do i=1,3
               do isig = -1,-1,2
                  
                  print*,"mat(",k,",",isig,",1,",i,") = ",mat(k,isig,1,i)
                  temp = dreal(mat(k,isig,1,i))**2+dimag(mat(k,isig,1,i))**2
                  print*,"m1kb(",1,",",isig,",",i,") = ",m1kb(1,isig,i)
                  temp1 = dreal(m1kb(1,isig,i))**2+dimag(m1kb(1,isig,i))**2
                  print*,"temp/temp1 = ",temp/temp1
                  
                  print*,"mat(",k,",",isig,",2,",i,") = ",mat(k,isig,2,i)
                  temp = dreal(mat(k,isig,2,i))**2+dimag(mat(k,isig,2,i))**2
                  print*,"m2kb(",1,",",isig,",",i,") = ",m2kb(1,isig,i)
                  temp1 = dreal(m2kb(1,isig,i))**2+dimag(m2kb(1,isig,i))**2
                  print*,"temp/temp1 = ",temp/temp1
               enddo
         enddo
         

      endif ! end of debugging block
c------------------------------------------------------
#endif

c sum the graphs, square them and map them onto uuwwaj and ddwwaj

      res(k) = 0
      do l = 1,2
         mm(k,-1,l) = 0
         do i = 1,3
            mm(k,-1,l) = mm(k,-1,l) +  mat(k,-1,l,i)
         enddo
         
         res(k)= res(k) + dreal(mm(k,-1,l))**2
     &        + dimag(mm(k,-1,l))**2
         
      enddo
      res(k) = res(k)*4d0*fpials(2) ! C_2*3 is the color factor
      
      final = res(k)
      
      return
      end


