***************************************************************************
*
*  This is the VBFNLO main program. 
*
****************************************************************************

****************************************************************************
*  Last modified: 03.05.2011
****************************************************************************

      PROGRAM vbfnlo
      implicit none

c include global variables via include-files, e.g. "procID"
#include "VBFNLO/utilities/global.inc"
#include "VBFNLO/utilities/lha.inc"
#include "VBFNLO/utilities/tau_hel.inc"
#include "VBFNLO/utilities/process.inc"

c define local variables
c--------------------------------------------------------
c     variables for the integration loops
      integer iteration, seed
      integer*8 Ncall, i, two
      parameter (two = 2)

c     digit characters for numbering the grid output file
      character*1 digit(61)
      data digit /'1','2','3','4','5','6','7','8','9','A',
     1            'B','C','D','E','F','G','H','I','J','K',
     1            'L','M','N','O','P','Q','R','S','T','U',
     1            'V','W','X','Y','Z','a','b','c','d','e',
     1            'f','g','h','i','j','k','l','m','n','o',
     1            'p','q','r','s','t','u','v','w','x','y',
     1            'z'/
c     the filenames for the grid output
      character*50 file_name_out, file_name_in

c     random number array controlled by monaco
      real*8 rand(max_PS_dim),rn

c     4-momenta of the partons involved in the basic QCD 2->2(+1)+X process.
c     The partons are the incoming quarks/gluons + the 2 tagging jets + 
c     the real emission (for NLO only)
      real*8 p(0:3,max_p,max_kin) 

c     Additional particles like a higgs boson, or any decay products 
c     (also hadronic ones). in short: everything which is not a parton 
c     involved in the basic QCD 2->2(+1)+X process.
      real*8 v(0:3,max_v,max_kin) 

c     Jets after jet-defining function
      real*8 jets(0:7,max_jets,max_kin) 
      integer n_jets(max_kin)
c     Charged leptons, i.e. electrons, muons, taus but not neutrinos
      real*8 leptons(0:8,max_v,max_kin)
      integer n_leptons(max_kin)
c     photons
      real*8 photons(0:7,max_v,max_kin)
      integer n_photons(max_kin)
c     invisible particles, i.e. neutrinos, LSP
      real*8 invisible(0:8,max_v,max_kin)
      integer n_invisible(max_kin)          
      
c     event weights, jacobi-factors, amplitude-squares
      real*8 weight, dps, m2s(0:max_kin), dsig(0:max_kin)

c     the result of the integration: crossection, error, chisquare
      real*8 chi2
      real*8 NLOxsection(-2:5), NLOsdev(-2:5), NLOchi2(-2:5)
c     crossection, error, chisquare for  real photon emission
      real*8 PHTNxsection, PHTNsdev, PHTNchi2
      real*8 ps_xsec(4), ps_sdev(4)

c     other variables
      real*8 x(nx)   ! feynman x parameters
      logical cut_ok ! whether event passes cuts
      logical ps_ok  ! phase space generator is suitable for this ps point
      real*8 xuz(2,2:max_kin)
      integer gnlo, L, mu, nd
      logical lokt(max_kin)
      common /nlovariable/ gnlo
      integer ps_number

c     external functions
      logical Cuts       ! general cut-function
      logical Choose_PS  ! general PS-function
      real*8 amplitude   ! general squared-amplitude function
      external Cuts, amplitude, Choose_PS

      real*8 RandomNumber
      external RandomNumber     

c     debugging and testing variables and switches
      logical showtiming
      parameter (showtiming=.false.)


      write(*,*)'  '
      write(*,*)'-----------------------------------------------------'
      write(*,*)'VBFNLO, version 2.5.0'
      write(*,*)'arXiV: 0411.xxxx'
      write(*,*)'http://www-itp.particle.uni-karlsruhe.de/~vbfnloweb'
      write(*,*)'-----------------------------------------------------'
      write(*,*)'  '
      write(*,*)'  '


** LoopTools initialisation
#ifdef WITH_LT
      call initLT      
#endif


      if (showtiming) call mtime

c     first, do all the initialization
    
      call InitGlobalParameters

      call PrintProcInfo

      call InitPDFs(0)

      call InitScales
      
      call InitCouplings

      call InitProcess

      call InitCuts
      
      call InitPhaseSpace
    
      call InitRandomNumbers(seed)

      call InitHistograms

c...Les Houches interface
      if (lha.and..not.doNLO) then 
         SELECT CASE (process)
         CASE(WMAJ, WPAJ, WPZJ, WMZJ,
     -        WMAJLO, WPAJLO, WPZJLO, WMZJLO )
            lha=.false.
            print*,''
            print*,'Sorry, LHA file is not available for the process',
     -           procID
            print*,''
         CASE DEFAULT
            lha=.true.
            call lha_file(-1)
         END SELECT
      endif
      

c==========================
c    The LO Calculation
c==========================
    
      do i=1,4
         ps_xsec(i)=0d0
         ps_sdev(i)=0
      enddo
      PHTNxsection=0d0
      PHTNsdev=0d0

      if (doBorn) then

         call InitPDFs(0)
         gnlo = 0

         do ps_number = 1,PS_Loops !multi-channel
            if (PS_Loops.gt.1) then
               write(*,*)" "
               write(*,*)"PHASE SPACE", ps_number
               write(*,*)" "
            endif

            file_name_in  = gridname1(ps_number)
     &                  (1:INDEX(gridname1(ps_number),'  ')-1)

            xsection=0d0
            sdev=0d0

            do iteration = 1, N_iterations1

               Select Case(procID)
               CASE(WPZA, WMZA)
                  if (ps_number.eq.1) then
                    Ncall = two**( N_points(0) - N_iterations1 + iteration )
                  else
                    Ncall = two**( N_points(0) + 1 - N_iterations1 + iteration )
                  endif
               CASE(WPAA, WMAA, ZAA)
                  if (ps_number.eq.1) then
                    Ncall = two**( N_points(0) +1 - N_iterations1 + iteration )
                  else if (ps_number.eq.4) then
                    Ncall = two**( N_points(0) -2 - N_iterations1 + iteration )
                  else
                    Ncall = two**( N_points(0)  - N_iterations1 + iteration )
                  endif
               CASE DEFAULT
                  Ncall = two**( N_points(0) - N_iterations1 + iteration )
               END SELECT
           
               file_name_out = gridname1(ps_number)
     &                    (1:INDEX(gridname1(ps_number),'  ')-1)//
     &                   '.out.'//digit(iteration)

               if ( iteration .eq. 1 ) then
                  call monaco_init( PS_dimension, Ncall, seed )
                  call monaco_read(file_name_in)
               else
                  call monaco_init2( PS_dimension, Ncall )
               end if       
               if (iteration.eq.N_iterations1) then
                  call monran_set(1) ! save random number info
               endif

               call ResetWeights

               do i = 1, Ncall
c     retrieve a set of random numbers and the corresponding weight
                  call monaco_get( rand, weight)

c     transform the random numbers to a physical phase space point
                  call phasespace(rand, p, x, v, ps_number, dps  )

                  if ( dps .gt. 0d0 ) then

c     define jets, leptons, photons
                     call defineJets(p, v, 1, jets, n_jets)
                     call DefineLeptons(v, 1, leptons, n_leptons)
                     call DefineInvisible(v,1, invisible, n_invisible)
                     call DefinePhotons(v, 1, photons, n_photons)

c in case of multi-channel: choose phase space:
                     PS_OK = Choose_PS(PS_number, v, 1)

c     apply cuts
                     cut_ok = Cuts(jets, n_jets, leptons, n_leptons, 
     1                              invisible,n_invisible,photons, n_photons, 1)

                      if (cut_ok .and. PS_ok) then
c     calculate fac. and ren. scale for this phase space point
                        call Scales(p, v, jets, n_jets, leptons, 
     1                          n_leptons, photons, n_photons, 1, 0)
c     call the matrix element code and get an amplitude square
                        m2s(0) = Amplitude(rand, p, x, v, 0, ps_number) 
                     else
                        dps = 0d0
                        m2s(0) = 0d0
                     end if 
                     dsig(0) = m2s(0)*dps

                     if (cut_ok .and. PS_ok) then
                        eventweight=dsig(0)*weight

c     in the last iteration : 
                        if (iteration.eq.N_iterations1) then

c     pass event information to the histogram routine
                           call HistogramEvent(weight,dsig(0),dps,
     &                                rand, p(0,1,1), x, v(0,1,1), 
     &                                jets(0,1,1), n_jets(1), 
     &                                leptons(0,1,1), n_leptons(1), 
     &                                invisible(0,1,1),n_invisible(1),
     &                                photons(0,1,1), n_photons(1), 0)

c     the unweighting of events : 
                       
c                          if (unweighting) then 
                           call Reweight(eventweight) ! reweight
c                          endif

                           if (eventweight.gt.0d0) then ! event is to be kept
c Les Houches interface ----------------
                              if((lha.or.hepmc).and..not.doNLO) call lha_put(p,v)

                           endif

                        endif
                     
                     endif ! cutok

                  else     ! dps > 0d0
                     m2s(0) = 0d0
                     dsig(0) = 0d0
                  end if

c     pass weight to monaco
                  call monaco_put(rand, weight, dsig(0))

               end do
        
               call monaco_write( file_name_out )

            enddo

            call monaco_result(xsection, sdev, chi2)
     
c     print out the result. 
            print *,""
            print *," result (LO): ",xsection," +- ",sdev," fb  "
     &           ,sdev/xsection*100d0,'%'
            print *,""
            ps_xsec(1)=ps_xsec(1)+xsection
            ps_sdev(1)=ps_sdev(1)+sdev**2
         enddo                  !multi-channel
         

         if (ps_loops.gt.1) then
            print *,""
            print *," Total (multi-channel) result (LO): ",
     &        ps_xsec(1)," +- ",sqrt(ps_sdev(1))," fb  ",
     &           sqrt(ps_sdev(1))/ps_xsec(1)*100d0,'%'
            print *," "
         endif

      else
         print *," No LO calculation performed."
      endif

      if (ps_xsec(1) .ne. ps_xsec(1)) then
         write(*,*)'Sorry, something has gone wrong with the LO'
         write(*,*)'cross section!'
         stop
      end if

c...Les Houches interface
      if(lha.and..not.doNLO) then
         xsection = ps_xsec(1)
         sdev = sqrt(ps_sdev(1))
         call lha_file(0)
         call lha_file(1)
      endif

      if(.not.doNLO) then 
         print*,'number of events in the last iteration =',ncall
         call reweightinginfo
      endif


#ifdef WITH_NLO
c============================
c    The NLO Calculation
c============================     
c  The virtual corrections      
c----------------------------
      if ((doVirtuals.and.hasNLO) .or. ewcor_switch) then

         call InitPDFs(1)

         do ps_number = 1,PS_Loops !multi-channel

            if (PS_Loops.gt.1) then
               print*," "
               print*,"PHASE SPACE ", ps_number
               print*," "
            endif

            pentcountps(ps_number)=0
            pentcount2ps(ps_number)=0

            if (N_iterations1.eq.1) then
               file_name_in  = gridname1(ps_number)
     &                 (1:INDEX(gridname1(ps_number),'  ')-1)
            else
               file_name_in  = gridname1(ps_number)
     &              (1:INDEX(gridname1(ps_number),'  ')-1)//
     &              '.out.'//digit(N_iterations1)
            endif

            NLOxsection(0) = 0d0
            NLOsdev(0) = 0d0

            do gnlo = 1, NLO_Loops ! loop over different virtual contributions

               Select Case(procID)
               CASE(WPZA, WMZA)
                  if (ps_number.eq.1) then
                    Ncall = two**(N_points(gnlo))
                  else
                    Ncall = two**(N_points(gnlo)+1)
                  endif
               CASE(WPAA, WMAA, ZAA)
                  if (ps_number.eq.1) then
                    Ncall = two**(N_points(gnlo)+1)
                  else if (ps_number.eq.4) then
                    Ncall = two**(N_points(gnlo)-2)
                  else
                    Ncall = two**(N_points(gnlo))
                  endif
               CASE DEFAULT
                  Ncall = two**(N_points(gnlo))
               END SELECT

               file_name_out = gridname1(ps_number)
     &                (1:INDEX(gridname1(ps_number),'  ')-1)//
     &                '.out.'//digit(1)
            
               call monaco_init( PS_dimension, Ncall, seed )
               call monaco_read(file_name_in)
!               call monran_set(2)  
         
               do i = 1, Ncall
                  call monaco_get( rand, weight)
         
                  call phasespace(rand, p, x, v, ps_number, dps  )

                  if ( dps .gt. 0d0 ) then         
               
                     call defineJets(p, v, 1, jets, n_jets)
                     call DefineLeptons(v, 1, leptons, n_leptons)
                     call DefinePhotons(v, 1, photons, n_photons)
                     call DefineInvisible(v,1, invisible, n_invisible)	  
                     
                     PS_OK = Choose_PS(PS_number, v, 1)

                     cut_ok = Cuts(jets, n_jets, leptons, n_leptons, 
     1                              invisible,n_invisible,photons, n_photons, 1)

                     if (cut_ok.and.ps_ok) then
                        call Scales(p, v, jets, n_jets, leptons, 
     1                             n_leptons, photons, n_photons, 1, 1)
                        m2s(0) = Amplitude(rand, p, x, v, gnlo, ps_number)
#ifdef WITH_LT  
                     call clearLT
#endif
                    else
                        dps = 0d0
                        m2s(0) = 0d0
                     end if 

                     dsig(0) = m2s(0)*dps

                     call HistogramEvent(weight,dsig(0),dps,
     &                                rand, p(0,1,1), x, v(0,1,1), 
     &                                jets(0,1,1), n_jets(1), 
     &                                leptons(0,1,1), n_leptons(1), 
     &                                invisible(0,1,1),n_invisible(1),     
     &                                photons(0,1,1), n_photons(1), 1)

                  else
                     m2s(0) = 0d0
                     dsig(0) = 0d0
                  end if

                  call monaco_put(rand, weight, dsig(0))
               end do

               call monaco_write(file_name_out)
     
               call monaco_result(NLOxsection(gnlo), 
     &                         NLOsdev(gnlo), NLOchi2(gnlo))

               print *," "
               print *," result for virtual contribution, no. ",gnlo
               print *," sigma = ",NLOxsection(gnlo)," +- ",
     &                 NLOsdev(gnlo)," fb  "

               Select Case(procID)
               CASE(WPZA,WMZA,WPAA,WMAA,ZAA,ZnAA,AAA)
                 print *," Pentagon instabilites", pentcountps(ps_number)
                 print *,"          total calls", 
     &                   pentcountps(ps_number)+pentcount2ps(ps_number)
                 if (Dble(pentcountps(ps_number)+pentcount2ps(ps_number)).ne.0d0) then
                   print *,"          instabilities [%]", 
     &                     pentcountps(ps_number)*100d0
     &                       /Dble(pentcountps(ps_number)+pentcount2ps(ps_number))
                 else
                   print *,"          instabilities [%]", 0d0
                 endif
               END SELECT

               print *," "
               NLOxsection(0) = NLOxsection(0) + NLOxsection(gnlo)
               NLOsdev(0) = sqrt(NLOsdev(0)**2 + NLOsdev(gnlo)**2)

               print *,' total result for virtual contributions so far '

               print *,' sigma = ',NLOxsection(0),' +- ',NLOsdev(0),
     -              ' fb'
               print *,' '

            enddo

            ps_xsec(2)=ps_xsec(2)+NLOxsection(0)
            ps_sdev(2)=ps_sdev(2)+NLOsdev(0)**2
         enddo !multi-channel

         if (ps_loops.gt.1) then
            print *,""
            print *," Total (multi-channel) result (NLO virtual): "
     &        ,ps_xsec(2)," +- ",sqrt(ps_sdev(2))," fb  "
     &        ,sqrt(ps_sdev(2))/ps_xsec(2)*100d0,'%'
            print *,""
         endif

      endif


      if (ps_xsec(2) .ne. ps_xsec(2)) then
         write(*,*)'Sorry, something has gone wrong with the NLO'
         write(*,*)'virtual result!'
         stop
      end if


** closing LoopTools
#ifdef WITH_LT
      call closeLT
#endif


c----------------------------
c    The real emissions      
c----------------------------
      if (doEmissions.and.hasNLO) then
         call InitPDFs(1)
         call InitRealEmissions

         do ps_number = 1,PS_Loops

            if (ps_loops.gt.1) then
               print*," "
               print*,"PHASE SPACE ", ps_number
               print*," "
            endif

            file_name_in  = gridname2(ps_number)
     &                    (1:INDEX(gridname2(ps_number),'  ')-1)

            NLOxsection(-1) = 0d0
            NLOsdev(-1) = 0d0

            DO iteration = 1, N_iterations2
               Select Case(procID)
               CASE(WPZA, WMZA)
                  if (ps_number.eq.1) then
                    Ncall = two**( N_points(-1) - N_iterations2 + iteration )
                  else
                    Ncall = two**( N_points(-1)+1 - N_iterations2 + iteration )
                  endif
               CASE(WPAA, WMAA, ZAA)
                  if (ps_number.eq.1) then
                    Ncall = two**( N_points(-1)+ 1 - N_iterations2 + iteration )
                  else if (ps_number.eq.4) then
                    Ncall = two**( N_points(-1)- 2 - N_iterations2 + iteration )
                  else
                    Ncall = two**( N_points(-1) - N_iterations2 + iteration )
                  endif
               CASE DEFAULT
                  Ncall = two**( N_points(-1) - N_iterations2 + iteration )
               END SELECT
           
               file_name_out = gridname2(ps_number)
     &               (1:INDEX(gridname2(ps_number),'  ')-1)//
     &               '.out.'//digit(iteration)
               if ( iteration .eq. 1 ) then
                  call monaco_init( PS_dimension, Ncall, seed )
                  call monaco_read(file_name_in)
               else
                  call monaco_init2( PS_dimension, Ncall )
               end if       
           
               do i = 1, Ncall
                  call monaco_get( rand, weight)

                  call phasespace(rand, p, x, v, ps_number, dps  )

                  if ( dps .gt. 0d0 ) then      
                     call ptilde(p(0,1,1),xuz(1,2),v(0,1,1))
                     call defineJets(p, v, n_kin, jets, n_jets)
                     call DefineLeptons(v, n_kin, leptons, n_leptons)
                     call DefinePhotons(v, n_kin, photons, n_photons)
                     call DefineInvisible(v,n_kin, invisible, n_invisible)

                     ps_ok = Choose_PS(PS_number, v, 1)                   
                                        
                     lokt(1) = Cuts(jets, n_jets, leptons, n_leptons,
     -                    invisible,n_invisible,photons,n_photons, 1)
     -                    .and. ps_ok

                     cut_ok = lokt(1)

                     do nd = 2, n_kin
                        ps_ok = Choose_PS(PS_number, v, nd)
                        lokt(nd) = Cuts(jets, n_jets, leptons, n_leptons,
     1                              invisible,n_invisible,
     1                              photons, n_photons, nd).and.ps_ok
                        cut_ok = cut_ok .or. lokt(nd)
                     enddo

                     if (cut_ok) then
                        call Scales(p, v, jets, n_jets, leptons, 
     1                          n_leptons, photons, n_photons, n_kin, 1)
                        call RE_Amplitude(rand, p, x, v, lokt, xuz, m2s,
     1                                                          ps_number)
                     else
                        dps = 0d0
                        do L = 0,n_kin
                           m2s(L) = 0d0
                        enddo                     
                     end if 

                     do L = 0,n_kin
                        dsig(L) = m2s(L)*dps
                     enddo   

                     if (iteration.eq.N_iterations2) then

                        if (dsig(0).ne.0d0) then
                           call HistogramEvent(weight,dsig(0),dps,
     &                                rand, p(0,1,1), x, v(0,1,1), 
     &                                jets(0,1,1), n_jets(1), 
     &                                leptons(0,1,1), n_leptons(1), 
     &                                invisible(0,1,1),n_invisible(1),
     &                                photons(0,1,1), n_photons(1), -1)
                        endif
                        if (dsig(0).ne.0d0.or.dsig(1).ne.0) then
                           call HistogramEvent(weight,dsig(1),dps,
     &                                rand, p(0,1,1), x, v(0,1,1), 
     &                                jets(0,1,1), n_jets(1), 
     &                                leptons(0,1,1), n_leptons(1), 
     &                                invisible(0,1,1),n_invisible(1),
     &                                photons(0,1,1), n_photons(1), 2)
                        endif
                        do L = 2,n_kin
                           if (dsig(L).ne.0d0) then
                              call HistogramEvent(weight,dsig(L),dps,
     &                                rand, p(0,1,L), x, v(0,1,L), 
     &                                jets(0,1,L), n_jets(L), 
     &                                leptons(0,1,L), n_leptons(L),
     &                                invisible(0,1,L),n_invisible(L),        
     &                                photons(0,1,L), n_photons(L), 2)

                           endif
                        enddo

                     endif

                  else
                     do L = 0,n_kin
                        m2s(L) = 0d0
                        dsig(L) = 0d0
                     enddo
                  end if   

                  call monaco_put(rand, weight, dsig(0))
               
               end do
        
               call monaco_write( file_name_out )

            END DO

            call monaco_result(NLOxsection(-1), NLOsdev(-1), NLOchi2(-1))
     
c     print out the result. 
            print *,""
            print *," result for real emissions "
            print *," sigma = ",NLOxsection(-1)," +- ",NLOsdev(-1)," fb"
            print *," "

            ps_xsec(3)=ps_xsec(3)+NLOxsection(-1)
            ps_sdev(3)=ps_sdev(3)+NLOsdev(-1)**2
         enddo                  !multi-channel

         if (ps_loops.gt.1) then
            print *,""
            print *," Total (multi-channel) result (NLO real): "
     &          ,ps_xsec(3)," +- ",sqrt(ps_sdev(3))," fb  "
     &          ,sqrt(ps_sdev(3))/ps_xsec(3)*100d0,'%'
            print *,""
         endif

         endif


*sophy: calculating the real photon emission.  Adapted from code written by
*       Terrance Figy
c----------------------------
c    The real photon emissions      
c----------------------------
      if (ewcor_switch .and. (sector .ge. 3)) then

         print *,""
         print *,"Starting real photon emissions"

         if (.not. doEmissions) call InitPDFs(1)

         call InitQEDEmissions

         do ps_number = 1,PS_Loops

            if (ps_loops.gt.1) then
               print*," "
               print*,"PHASE SPACE ", ps_number
               print*," "
            endif

            file_name_in  = gridname3(ps_number)
     &                    (1:INDEX(gridname3(ps_number),'  ')-1)

            PHTNxsection=0d0
            PHTNsdev=0d0

            DO iteration = 1, N_iterations2
               Ncall = 2**( N_points(-1) - N_iterations2 + iteration )
              
               file_name_out = gridname3(ps_number)
     &               (1:INDEX(gridname3(ps_number),'  ')-1)//
     &               '.out.'//digit(iteration)
               if ( iteration .eq. 1 ) then
                  call monaco_init( PS_dimension, Ncall, seed )
                  call monaco_read(file_name_in)
               else
                  call monaco_init2( PS_dimension, Ncall )
               end if       
           
               do i = 1, Ncall
                  call monaco_get( rand, weight)

                  call phasespace(rand, p, x, v, ps_number, dps  )

                  if ( dps .gt. 0d0 ) then      
                     call ptilde_qed(p(0,1,1),xuz(1,2),v(0,1,1)) 
                     call defineJets(p, v, n_qed, jets, n_jets)
                     call DefineLeptons(v, n_qed, leptons, n_leptons)
                     call DefinePhotons(v, n_qed, photons, n_photons)
                     call DefineInvisible(v,n_qed, invisible, n_invisible)
             
                     ps_ok = Choose_PS(PS_number, v, 1)
                     
                     lokt(1) = Cuts(jets, n_jets, leptons, n_leptons,
     1                                            invisible,n_invisible,
     1                                            photons,n_photons, 1)
     1                         .and.ps_ok
                     cut_ok = lokt(1)
                     do nd = 2, n_qed
                        ps_ok = Choose_PS(PS_number, v, nd)
                        lokt(nd) = Cuts(jets, n_jets, leptons, n_leptons,
     1                                              invisible,n_invisible,
     1                                              photons, n_photons, nd)
     1                         .and.ps_ok
                        cut_ok = cut_ok .or. lokt(nd)
                     enddo
                     if (cut_ok) then
                        call Scales(p, v, jets, n_jets, leptons, 
     1                             n_leptons, photons, n_photons, n_kin, 1)
                        call QED_Amplitude(rand, p, x, v, lokt, xuz, m2s,
     1                                                          ps_number)
                     else
                        dps = 0d0
                        do L = 0,n_qed
                           m2s(L) = 0d0
                        enddo                     
                     end if 

                     do L = 0,n_qed
                        dsig(L) = m2s(L)*dps
                     enddo   

                     if (iteration.eq.N_iterations2) then

                           if (dsig(0).ne.0d0) then
                              call HistogramEvent(weight,dsig(0),dps,
     &                                   rand, p(0,1,1), x, v(0,1,1), 
     &                                   jets(0,1,1), n_jets(1), 
     &                                   leptons(0,1,1), n_leptons(1), 
     &                                   invisible(0,1,1),n_invisible(1),
     &                                   photons(0,1,1), n_photons(1), -1)
                           endif
                           if (dsig(0).ne.0d0.or.dsig(1).ne.0) then
                              call HistogramEvent(weight,dsig(1),dps,
     &                                   rand, p(0,1,1), x, v(0,1,1), 
     &                                   jets(0,1,1), n_jets(1), 
     &                                   leptons(0,1,1), n_leptons(1), 
     &                                   invisible(0,1,1),n_invisible(1),
     &                                   photons(0,1,1), n_photons(1), 2)
                           endif
                           do L = 2,n_qed
                              if (dsig(L).ne.0d0) then
                                 call HistogramEvent(weight,dsig(L),dps,
     &                                   rand, p(0,1,L), x, v(0,1,L), 
     &                                   jets(0,1,L), n_jets(L), 
     &                                   leptons(0,1,L), n_leptons(L),
     &                                   invisible(0,1,L),n_invisible(L),   
     &                                   photons(0,1,L), n_photons(L), 2)
                           endif
                        enddo

                     endif

                  else
                     do L = 0,n_qed
                        m2s(L) = 0d0
                        dsig(L) = 0d0
                     enddo
                  end if
                  
                  call monaco_put(rand, weight, dsig(0))
                  
               end do
            
               call monaco_write( file_name_out )

            END DO

            call monaco_result(PHTNxsection, PHTNsdev, PHTNchi2)
     
c     print out the result. 
            print *,""
            print *," result for QED real emissions "
            print *," sigma = ",PHTNxsection," +- ",PHTNsdev," fb"
            print *," "

            ps_xsec(4)=ps_xsec(4)+PHTNxsection
            ps_sdev(4)=ps_sdev(4)+PHTNsdev**2
         enddo                  !multi-channel

         if (ps_loops.gt.1) then
            print *,""
            print *," Total (multi-channel) result (QED real): "
     &          ,ps_xsec(4)," +- ",sqrt(ps_sdev(4))," fb  "
     &          ,sqrt(ps_sdev(4))/ps_xsec(4)*100d0,'%'
            print *,""
         endif

      end if
      
         if ((dovirtuals.and.doEmissions.and.hasNLO).or.ewcor_switch) then
            NLOxsection(-2) = ps_xsec(2)+ps_xsec(3)+ps_xsec(4)
            NLOsdev(-2) = sqrt(ps_sdev(2)+ps_sdev(3)+ps_sdev(4))
            print *," "
            print *," final result at NLO "         
            print *," sigma = ",NLOxsection(-2)," +- ",NLOsdev(-2)," fb"
     &      ,NLOsdev(-2)/NLOxsection(-2)*100d0,'%'
         endif

#endif

      call WriteHistograms

     
c print out results in a file also
      open (20,
     &      file="xsection.out",
     &      status='REPLACE')
      write(20,*) ps_xsec(1), sqrt(ps_sdev(1))
      write(20,*) NLOxsection(-2), NLOsdev(-2)
      close(20)

      Select Case(procID)
      CASE(WPZA,WMZA,WPAA,WMAA,ZAA,ZnAA,AAA)
        print *," Pentagon instabilites", pentcountm
        print *,"          total calls", 
     &          pentcountm+pentcount2m
        if (Dble(pentcountm+pentcount2m).ne.0d0) then
          print *,"          instabilities [%]", 
     &          pentcountm/Dble(pentcountm+pentcount2m)*100d0
        else
          print *,"          instabilities [%]", 0d0
        endif
      END SELECT


      if (showtiming) call mtime

c...Les Houches interface
      if (lha.and..not.doNLO) then 
         print*,''
         print*,'created LHA file for the LO calculation'
         print*,''
      endif
      if(lha.and.doNLO) then
         print*,''
         print*,'No LHA file for the NLO calculation'
         print*,''
      endif
   
      STOP


      END
c     END OF VBFNLO
    


